from boa.interop.Neo.Storage import Get, Put, Delete, Find, GetContext
from boa.interop.Neo.Iterator import IterNext, IterKey, IterValue
from boa.interop.Neo.Runtime import Notify

ctx = GetContext()


def Main(query):

    Put(ctx, 'prefix1euo', 1)
    Put(ctx, 'prefix1e', 2)
    Put(ctx, 'prefix1__osetuh', 3)

    Put(ctx, 'blah', 'Hello Storage Find')

    result_iter = Find(ctx, query)

    items = []
    keys = []
    count = 0
    while result_iter.IterNext():
        val = result_iter.IterValue()
        items.append(val)
        keys.append(result_iter.IterKey())
        if query == 'pre' and count == 1:
            break

        count += 1

    if query == 'pref':
        return keys

    return items



from boa.interop.Neo.Runtime import Notify


def Main():

    m = [1, 2, 2]

    m.append(7)

    q = [6, 7]

    l = 'howdy'

    m.append(l)

    m.append(q)

    m.append(b'\x01')

    answer = q[0]

#    Notify(m)

    return m[5]



from boa_test.tests.boa_test import BoaTest
from boa.compiler import Compiler

from neo.Prompt.Commands.BuildNRun import TestBuild


class TestContract(BoaTest):

    def test_opcalls(self):

        output = Compiler.instance().load('%s/boa_test/example/OpCallTest.py' % TestContract.dirname).default
        out = output.write()

        tx, results, total_ops, engine = TestBuild(out, ['omin', 4, 4], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 4)

        tx, results, total_ops, engine = TestBuild(out, ['omin', -4, 4], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), -4)

        tx, results, total_ops, engine = TestBuild(out, ['omin', 16, 0], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 0)

        tx, results, total_ops, engine = TestBuild(out, ['omax', 4, 4], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 4)

        tx, results, total_ops, engine = TestBuild(out, ['omax', -4, 4], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 4)

        tx, results, total_ops, engine = TestBuild(out, ['omax', 16, 0], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 16)

        tx, results, total_ops, engine = TestBuild(out, ['oabs', 0, 4], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 0)

        tx, results, total_ops, engine = TestBuild(out, ['oabs', -4, 4], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 4)

        tx, results, total_ops, engine = TestBuild(out, ['sha1', 'abc', 4], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b'\xa9\x99>6G\x06\x81j\xba>%qxP\xc2l\x9c\xd0\xd8\x9d'))

        tx, results, total_ops, engine = TestBuild(out, ['sha256', 'abc', 4], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b'\xbax\x16\xbf\x8f\x01\xcf\xeaAA@\xde]\xae"#\xb0\x03a\xa3\x96\x17z\x9c\xb4\x10\xffa\xf2\x00\x15\xad'))

        tx, results, total_ops, engine = TestBuild(out, ['hash160', 'abc', 4], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b'\xbb\x1b\xe9\x8c\x14$D\xd7\xa5j\xa3\x98\x1c9B\xa9x\xe4\xdc3'))

        tx, results, total_ops, engine = TestBuild(out, ['hash256', 'abc', 4], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b'O\x8bB\xc2-\xd3r\x9bQ\x9b\xa6\xf6\x8d-\xa7\xcc[-`m\x05\xda\xedZ\xd5\x12\x8c\xc0>lcX'))



"""
Large Array Storage Test

A simple utilise contract that allows you to manipulate a stored list for stress tests and GAS cost evaluation.

Test Command:
    build [FILE_PATH] test 0710 05 True False init

Example Executions:
    testinvoke [CONTRACT_HASH] init
    testinvoke [CONTRACT_HASH] delete
    testinvoke [CONTRACT_HASH] fetch
    testinvoke [CONTRACT_HASH] count
    testinvoke [CONTRACT_HASH] append_1
    testinvoke [CONTRACT_HASH] append_10
"""
from boa.interop.Neo.Storage import Get, Put, Delete, GetContext
from boa.interop.Neo.Runtime import Log, Notify
from boa.builtins import concat, list, range, take, substr

# -- Global variables
KEY = 'test_array'


def Main(operation, args):
    """

    :param operation: The name of the operation to perform
    :param args: A list of arguments along with the operation
    :type operation: str
    :type args: list
    :return: The result of the operation
    :rtype: bytearray
    """
    if operation == 'init':
        return do_init()
    elif operation == 'delete':
        return do_delete()
    elif operation == 'fetch':
        return do_fetch()
    elif operation == 'count':
        return do_count()
    elif operation == 'append_1':
        return do_append_1()
    elif operation == 'append_10':
        return do_append_10()
    else:
        Notify('unknown operation')
        return False


def do_init():
    """
    Initialize the storage by setting an empty list.

    :return: indication success execution of the command
    :rtype: bool
    """
    context = GetContext()
    init_list_bytes = serialize_array([])
    Put(context, KEY, init_list_bytes)
    return True


def do_delete():
    """
    Delete the storage and reset back to its default state.

    :return: indication success execution of the command
    :rtype: bool
    """
    context = GetContext()
    Delete(context, KEY)
    return True


def do_fetch():
    """
    Fetch current list value from storage.

    :return: the stored list value
    :rtype: list
    """
    context = GetContext()
    list_bytes = Get(context, KEY)
    return deserialize_bytearray(list_bytes)


def do_count():
    """
    Fetch length of the stored list.

    :return: the stored list value
    :rtype: int
    """
    context = GetContext()
    list_bytes = Get(context, KEY)
    item_list = deserialize_bytearray(list_bytes)
    return len(item_list)


def do_append_1():
    """
    Add 1 item into the list in storage.

    :return: indication success execution of the command
    :rtype: bool
    """
    context = GetContext()
    list_bytes = Get(context, KEY)
    item_list = deserialize_bytearray(list_bytes)
    item_list.append('single item')
    list_length = len(item_list)
    Log('new list length:')
    Log(list_length)
    list_bytes = serialize_array(item_list)
    Put(context, KEY, list_bytes)
    return True


def do_append_10():
    """
    Add 10 items into the list in storage.

    :return: indication success execution of the command
    :rtype: bool
    """
    context = GetContext()
    list_bytes = Get(context, KEY)
    item_list = deserialize_bytearray(list_bytes)
    addition_list = [
        '1 of 10', '1 of 10', '1 of 10', '1 of 10', '1 of 10', '1 of 10', '1 of 10', '1 of 10', '1 of 10', '1 of 10'
    ]
    for item in addition_list:
        item_list.append(item)
    list_length = len(item_list)
    Log('new list length:')
    Log(list_length)
    list_bytes = serialize_array(item_list)
    Put(context, KEY, list_bytes)
    return True


def deserialize_bytearray(data):

    # get length of length
    collection_length_length = data[0:1]

    # get length of collection
    collection_len = data[1:collection_length_length + 1]

    # create a new collection
    new_collection = list(length=collection_len)

    # trim the length data
    offset = 1 + collection_length_length

    for i in range(0, collection_len):

        # get the data length length
        itemlen_len = data[offset:offset + 1]

        # get the length of the data
        item_len = data[offset + 1:offset + 1 + itemlen_len]

        # get the data
        item = data[offset + 1 + itemlen_len: offset + 1 + itemlen_len + item_len]

        # store it in collection
        new_collection[i] = item

        offset = offset + item_len + itemlen_len + 1

    return new_collection


def serialize_array(items):

    # serialize the length of the list
    itemlength = serialize_var_length_item(items)

    output = itemlength

    # now go through and append all your stuff
    for item in items:

        # get the variable length of the item
        # to be serialized
        itemlen = serialize_var_length_item(item)

        # add that indicator
        output = concat(output, itemlen)

        # now add the item
        output = concat(output, item)

    # return the stuff
    return output


def serialize_var_length_item(item):

    # get the length of your stuff
    stuff_len = len(item)

    # now we need to know how many bytes the length of the array
    # will take to store

    # this is one byte
    if stuff_len <= 255:
        byte_len = b'\x01'
    # two byte
    elif stuff_len <= 65535:
        byte_len = b'\x02'
    # hopefully 4 byte
    else:
        byte_len = b'\x04'

    out = concat(byte_len, stuff_len)

    return out



# tested

from boa.interop.Neo.Blockchain import GetTransaction
from boa.interop.Neo.Transaction import *
from boa.interop.Neo.Output import *

NEO = b'\x9b|\xff\xda\xa6t\xbe\xae\x0f\x93\x0e\xbe`\x85\xaf\x90\x93\xe5\xfeV\xb3J\\"\x0c\xcd\xcfn\xfc3o\xc5'
GAS = b'\xe7-(iy\xeel\xb1\xb7\xe6]\xfd\xdf\xb2\xe3\x84\x10\x0b\x8d\x14\x8ewX\xdeB\xe4\x16\x8bqy,`'


def Main(txid):

    tx = GetTransaction(txid)

    res = []

    # we can iterate over an attribute of an object!
    # this is really exciting, but you'd never know it
    for item in tx.Outputs:
        subres = []
        subres.append(item.Value)
        subres.append(item.AssetId)
        subres.append(item.ScriptHash)
        res.append(subres)

    return res



# tested
from boa.builtins import list
from boa.interop.Neo.Runtime import Notify


def Main():

    m = 3

    j = list(length=m)

    j[0] = 3

    j[1] = 2

    q = j[0]

    Notify(q)

    return j



from boa.interop.Neo.App import RegisterAppCall

CalculatorContract = RegisterAppCall(b'W\xa7\x18\x08MZh\xbdu\xb7%\x88\x8e\x19J\x9e\xd4|\xe1\xe3\xaa', 'operation', 'a', 'b')


def Main(operation, a, b):

    result = CalculatorContract(operation, a, b)

    return result



from boa_test.tests.boa_test import BoaTest
from boa.compiler import Compiler

from neo.Prompt.Commands.BuildNRun import TestBuild

from neo.Settings import settings


class TestContract(BoaTest):

    def test_ManyElif(self):

        output = Compiler.instance().load('%s/boa_test/example/TestManyElif.py' % TestContract.dirname).default
        out = output.write()

        tx, results, total_ops, engine = TestBuild(out, [1], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 2)

        tx, results, total_ops, engine = TestBuild(out, [3], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 4)

        tx, results, total_ops, engine = TestBuild(out, [16], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 17)

        tx, results, total_ops, engine = TestBuild(out, [22], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), -1)



# tested

from boa.interop.Neo.Blockchain import GetAccount
from boa.interop.Neo.Account import *

NEO = b'\x9b|\xff\xda\xa6t\xbe\xae\x0f\x93\x0e\xbe`\x85\xaf\x90\x93\xe5\xfeV\xb3J\\"\x0c\xcd\xcfn\xfc3o\xc5'
GAS = b'\xe7-(iy\xeel\xb1\xb7\xe6]\xfd\xdf\xb2\xe3\x84\x10\x0b\x8d\x14\x8ewX\xdeB\xe4\x16\x8bqy,`'


def Main(operation, acct):

    account = GetAccount(acct)

    if not account:
        return False

    if operation == 'get_hash':
        return account.ScriptHash

    elif operation == 'get_votes':
        return account.Votes

    elif operation == 'get_balance_gas':
        return GetBalance(account, GAS)

    elif operation == 'get_balance_neo':
        return GetBalance(account, NEO)

    return 'unknown operation'



from boa.interop.Neo.Runtime import CheckWitness, Notify
from boa.interop.Neo.Action import RegisterAction
from boa.interop.Neo.Storage import *
from boa.builtins import concat

from boa_test.example.demo.nex.token import *


OnTransfer = RegisterAction('transfer', 'addr_from', 'addr_to', 'amount')
OnApprove = RegisterAction('approve', 'addr_from', 'addr_to', 'amount')


def handle_nep51(ctx, operation, args):

    if operation == 'name':
        return TOKEN_NAME

    elif operation == 'decimals':
        return TOKEN_DECIMALS

    elif operation == 'symbol':
        return TOKEN_SYMBOL

    elif operation == 'totalSupply':
        return Get(ctx, TOKEN_CIRC_KEY)

    elif operation == 'balanceOf':
        if len(args) == 1:
            return Get(ctx, args[0])

    elif operation == 'transfer':
        if len(args) == 3:
            return do_transfer(ctx, args[0], args[1], args[2])

    elif operation == 'transferFrom':
        if len(args) == 3:
            return do_transfer_from(ctx, args[0], args[1], args[2])

    elif operation == 'approve':
        if len(args) == 3:
            return do_approve(ctx, args[0], args[1], args[2])

    elif operation == 'allowance':
        if len(args) == 2:
            return do_allowance(ctx, args[0], args[1])

    return False


def do_transfer(ctx, t_from, t_to, amount):

    if amount <= 0:
        return False

    if len(t_to) != 20:
        return False

    if CheckWitness(t_from):

        if t_from == t_to:
            print("transfer to self!")
            return True

        from_val = Get(ctx, t_from)

        if from_val < amount:
            print("insufficient funds")
            return False

        if from_val == amount:
            Delete(ctx, t_from)

        else:
            difference = from_val - amount
            Put(ctx, t_from, difference)

        to_value = Get(ctx, t_to)

        to_total = to_value + amount

        Put(ctx, t_to, to_total)

        OnTransfer(t_from, t_to, amount)

        return True
    else:
        print("from address is not the tx sender")

    return False


def do_transfer_from(ctx, t_from, t_to, amount):

    if amount <= 0:
        return False

    available_key = concat(t_from, t_to)

    if len(available_key) != 40:
        return False

    available_to_to_addr = Get(ctx, available_key)

    if available_to_to_addr < amount:
        print("Insufficient funds approved")
        return False

    from_balance = Get(ctx, t_from)

    if from_balance < amount:
        print("Insufficient tokens in from balance")
        return False

    to_balance = Get(ctx, t_to)

    new_from_balance = from_balance - amount

    new_to_balance = to_balance + amount

    Put(ctx, t_to, new_to_balance)
    Put(ctx, t_from, new_from_balance)

    print("transfer complete")

    new_allowance = available_to_to_addr - amount

    if new_allowance == 0:
        print("removing all balance")
        Delete(ctx, available_key)
    else:
        print("updating allowance to new allowance")
        Put(ctx, available_key, new_allowance)

    OnTransfer(t_from, t_to, amount)

    return True


def do_approve(ctx, t_owner, t_spender, amount):

    if not CheckWitness(t_owner):
        return False

    if amount < 0:
        return False

    # cannot approve an amount that is
    # currently greater than the from balance
    if Get(ctx, t_owner) >= amount:

        approval_key = concat(t_owner, t_spender)

        if amount == 0:
            Delete(ctx, approval_key)
        else:
            Put(ctx, approval_key, amount)

        OnApprove(t_owner, t_spender, amount)

        return True

    return False


def do_allowance(ctx, t_owner, t_spender):

    return Get(ctx, concat(t_owner, t_spender))



from boa_test.tests.boa_test import BoaTest
from boa.compiler import Compiler

from neo.Prompt.Commands.BuildNRun import TestBuild


class TestContract(BoaTest):

    def test_CompareTest0(self):

        output = Compiler.instance().load('%s/boa_test/example/CompareTest0.py' % TestContract.dirname).default
        out = output.write()

        tx, results, total_ops, engine = TestBuild(out, [2, 4], self.GetWallet1(), '0202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 2)

        tx, results, total_ops, engine = TestBuild(out, [4, 2], self.GetWallet1(), '0202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 3)

        tx, results, total_ops, engine = TestBuild(out, [2, 2], self.GetWallet1(), '0202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 2)

        tx, results, total_ops, engine = TestBuild(out, ['b', 'a'], self.GetWallet1(), '0202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 3)

        tx, results, total_ops, engine = TestBuild(out, ['a', 'b'], self.GetWallet1(), '0202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 2)

    def test_CompareTest1(self):

        output = Compiler.instance().load('%s/boa_test/example/CompareTest1.py' % TestContract.dirname).default
        out = output.write()

        tx, results, total_ops, engine = TestBuild(out, [1, 2, 3, 4], self.GetWallet1(), '02020202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 11)

        tx, results, total_ops, engine = TestBuild(out, [1, 2, 4, 3], self.GetWallet1(), '02020202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 1)

        tx, results, total_ops, engine = TestBuild(out, [1, 4, 3, 5], self.GetWallet1(), '02020202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 22)

        tx, results, total_ops, engine = TestBuild(out, [4, 1, 5, 3], self.GetWallet1(), '02020202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 3)

        tx, results, total_ops, engine = TestBuild(out, [9, 1, 3, 5], self.GetWallet1(), '02020202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 10)

        tx, results, total_ops, engine = TestBuild(out, [9, 5, 3, 5], self.GetWallet1(), '02020202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 8)

    def test_CompareTest2(self):

        output = Compiler.instance().load('%s/boa_test/example/CompareTest2.py' % TestContract.dirname).default
        out = output.write()

        tx, results, total_ops, engine = TestBuild(out, [2, 2], self.GetWallet1(), '0202', '01')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBoolean(), True)

        tx, results, total_ops, engine = TestBuild(out, [2, 3], self.GetWallet1(), '0202', '01')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBoolean(), False)



from boa_test.tests.boa_test import BoaFixtureTest
from boa.compiler import Compiler
from neo.Core.TX.Transaction import Transaction
from neo.Prompt.Commands.BuildNRun import TestBuild


class TestContract(BoaFixtureTest):

    def test_Appcall1(self):

        with self.assertRaises(Exception):
            Compiler.instance().load('%s/boa_test/example/blockchain/AppCallTest2.py' % TestContract.dirname).default

        with self.assertRaises(Exception):
            Compiler.instance().load('%s/boa_test/example/blockchain/AppCallTest3.py' % TestContract.dirname).default

    def test_Appcall2(self):

        output = Compiler.instance().load('%s/boa_test/example/blockchain/AppCallTest.py' % TestContract.dirname).default
        out = output.write()

        tx, results, total_ops, engine = TestBuild(out, ['add', 3, 5], self.GetWallet1(), '070202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 8)

        tx, results, total_ops, engine = TestBuild(out, ['sub', 3, 5], self.GetWallet1(), '070202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), -2)

        tx, results, total_ops, engine = TestBuild(out, ['mul', 3, 5], self.GetWallet1(), '070202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 15)

        tx, results, total_ops, engine = TestBuild(out, ['notfound', 3, 5], self.GetWallet1(), '070202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 0)

    def test_DynamicAppcall(self):

        output = Compiler.instance().load('%s/boa_test/example/blockchain/DynamicAppCallTest.py' % TestContract.dirname).default
        out = output.write()

        tx, results, total_ops, engine = TestBuild(out, [bytearray(b'W\xa7\x18\x08MZh\xbdu\xb7%\x88\x8e\x19J\x9e\xd4|\xe1\xe9'), 'add', 3, 5], self.GetWallet1(), '05070202', '02', dynamic=True)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 8)

        tx, results, total_ops, engine = TestBuild(out, [bytearray(b'W\xa7\x18\x08MZh\xbdu\xb7%\x88\x8e\x19J\x9e\xd4|\xe1\xe9'), 'sub', 3, 5], self.GetWallet1(), '05070202', '02', dynamic=True)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), -2)

        tx, results, total_ops, engine = TestBuild(out, [bytearray(b'W\xa7\x18\x08MZh\xbdu\xb7%\x88\x8e\x19J\x9e\xd4|\xe1\xe9'), 'mul', 3, 5], self.GetWallet1(), '05070202', '02', dynamic=True)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 15)

        tx, results, total_ops, engine = TestBuild(out, [bytearray(b'W\xa7\x18\x08MZh\xbdu\xb7%\x88\x8e\x19J\x9e\xd4|\xe1\xe3'), 'add', 3, 5], self.GetWallet1(), '05070202', '02', dynamic=True)
        self.assertEqual(len(results), 0)
#        self.assertEqual(results[0].GetBigInteger(), 0)



from boa_test.tests.boa_test import BoaTest
from boa.compiler import Compiler

from neo.Prompt.Commands.BuildNRun import TestBuild


class TestContract(BoaTest):

    def test_Slice1(self):
        output = Compiler.instance().load('%s/boa_test/example/SliceTest.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b'\x01\x02\x03\x04'))

    def test_Slice2(self):
        output = Compiler.instance().load('%s/boa_test/example/SliceTest2.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b'\x02\x03\x04\x02\x03\x04\x05\x06\x01\x02\x03\x04\x03\x04'))



from boa_test.tests.boa_test import BoaTest
from boa.compiler import Compiler

from neo.Prompt.Commands.BuildNRun import TestBuild
from neo.EventHub import events, SmartContractEvent


class TestContract(BoaTest):

    def test_Event1(self):

        dispatched_events = []

        def on_notif(evt):
            dispatched_events.append(evt)
        events.on(SmartContractEvent.RUNTIME_NOTIFY, on_notif)

        output = Compiler.instance().load('%s/boa_test/example/blockchain/EventTest.py' % TestContract.dirname).default
        out = output.write()

        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)

        self.assertEqual(results[0].GetBigInteger(), 7)

        events.off(SmartContractEvent.RUNTIME_NOTIFY, on_notif)

        self.assertEqual(len(dispatched_events), 2)

        transfer_event_payload = dispatched_events[0].event_payload

        self.assertEqual(len(transfer_event_payload), 4)
        self.assertEqual(transfer_event_payload[0], b'transfer_test')
        self.assertEqual(transfer_event_payload[1], 2)
        self.assertEqual(transfer_event_payload[2], 5)
        self.assertEqual(transfer_event_payload[3], 7)

        refund_event_payload = dispatched_events[1].event_payload
        self.assertEqual(len(refund_event_payload), 3)
        self.assertEqual(refund_event_payload[0], b'refund')
        self.assertEqual(refund_event_payload[1], b'me')
        self.assertEqual(int.from_bytes(refund_event_payload[2], 'little'), 52)



# tested

from boa.builtins import range


def Main():

    rangestart = 2
    count = 0

    for i in range(rangestart, getrangeend()):
        count += 1

    return count


def getrangeend():
    return 8



from boa.interop.System.ExecutionEngine import GetScriptContainer, GetExecutingScriptHash
from boa.interop.Neo.Transaction import Transaction, GetReferences, GetOutputs, GetUnspentCoins
from boa.interop.Neo.Output import GetValue, GetAssetId, GetScriptHash


neo_asset_id = b'\x9b|\xff\xda\xa6t\xbe\xae\x0f\x93\x0e\xbe`\x85\xaf\x90\x93\xe5\xfeV\xb3J\\"\x0c\xcd\xcfn\xfc3o\xc5'

gas_asset_id = b'\xe7-(iy\xeel\xb1\xb7\xe6]\xfd\xdf\xb2\xe3\x84\x10\x0b\x8d\x14\x8ewX\xdeB\xe4\x16\x8bqy,`'


def get_asset_attachments():
    """
    Gets information about NEO and Gas attached to an invocation TX

    :return:
        list: A list with information about attached neo and gas
    """

    tx = GetScriptContainer()
    references = tx.References

    receiver_addr = GetExecutingScriptHash()
    sender_addr = None
    sent_amount_neo = 0
    sent_amount_gas = 0

    if len(references) > 0:

        reference = references[0]
        sender_addr = reference.ScriptHash
        for output in tx.Outputs:
            if output.ScriptHash == receiver_addr:
                if output.AssetId == neo_asset_id:
                    sent_amount_neo += output.Value
                if output.AssetId == gas_asset_id:
                    sent_amount_gas += output.Value

    return [receiver_addr, sender_addr, sent_amount_neo, sent_amount_gas]



from boa.interop.Neo.Header import GetIndex, GetHash, GetPrevHash, GetTimestamp, GetVersion, GetNextConsensus, GetMerkleRoot, GetConsensusData


class Block:

    @property
    def TransactionCount(self):
        """

        :return:
        """
        return GetTransactionCount(self)

    @property
    def Transactions(self):
        """

        :return:
        """
        return GetTransactions(self)

    @property
    def Index(self):
        """

        Returns:

        """
        return GetIndex(self)

    @property
    def Hash(self):
        """

        :return:
        """
        return GetHash(self)

    @property
    def Timestamp(self):
        """

        :return:
        """
        return GetTimestamp(self)

    @property
    def Version(self):
        """

        :return:
        """
        return GetVersion(self)

    @property
    def PrevHash(self):
        """

        :return:
        """
        return GetPrevHash(self)

    @property
    def MerkleRoot(self):
        """

        :return:
        """
        return GetMerkleRoot(self)

    @property
    def ConsensusData(self):
        """

        :return:
        """
        return GetConsensusData(self)

    @property
    def NextConsensus(self):
        """

        :return:
        """
        return GetNextConsensus(self)


def GetTransactionCount(block):
    """

    returns the number of transactions in a block


    """
    pass


def GetTransactions(block):
    """
    returns a list of transactions contained in a block


    """
    pass


def GetTransaction(block, index):
    """

    :param block: the block to get the transaction from
    :param index: the index of the transaction within the lock


    """

    pass



from boa_test.tests.boa_test import BoaTest
from boa.compiler import Compiler
from neo.Settings import settings
from neo.Prompt.Commands.BuildNRun import TestBuild

# settings.set_log_vm_instruction(True)
# settings.log_smart_contract_events = True


class TestContract(BoaTest):

    def test_dict1(self):

        output = Compiler.instance().load('%s/boa_test/example/DictTest4.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 10)

    def test_dict2(self):

        with self.assertRaises(Exception) as ctx:
            Compiler.instance().load('%s/boa_test/example/DictTest5_ShouldNotCompile.py' % TestContract.dirname).default.write()

    def test_dict3(self):

        with self.assertRaises(Exception) as ctx:
            Compiler.instance().load('%s/boa_test/example/DictTest6_ShouldNotCompile.py' % TestContract.dirname).default.write()

    def test_dict_keys1(self):
        output = Compiler.instance().load('%s/boa_test/example/DictTestKeys.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b'ab\x04mzmcallltrs'))

    def test_dict_values1(self):
        output = Compiler.instance().load('%s/boa_test/example/DictTestValues.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 55)

    def test_dict_has_key(self):
        output = Compiler.instance().load('%s/boa_test/example/DictTestHasKey.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 22)



from neo.Implementations.Wallets.peewee.UserWallet import UserWallet
from neo.Settings import settings
from neocore.UInt160 import UInt160
from neo.Wallets.utils import to_aes_key
import os
from neo.Utils.WalletFixtureTestCase import WalletFixtureTestCase

settings.USE_DEBUG_STORAGE = False
# settings.DEBUG_STORAGE_PATH = './fixtures/debugstorage'


class BoaTest(WalletFixtureTestCase):

    dirname = None

    FIXTURE_REMOTE_LOC = 'https://s3.us-east-2.amazonaws.com/cityofzion/fixtures/fixtures_v7.tar.gz'
    FIXTURE_FILENAME = 'fixtures/empty_fixture.tar.gz'

    wallet_1_script_hash = UInt160(data=b'S\xefB\xc8\xdf!^\xbeZ|z\xe8\x01\xcb\xc3\xac/\xacI)')

    wallet_1_addr = 'APRgMZHZubii29UXF9uFa6sohrsYupNAvx'

    _wallet1 = None

    @classmethod
    def setUpClass(cls):

        cls.dirname = '/'.join(os.path.abspath(__file__).split('/')[:-3])

        super(BoaTest, cls).setUpClass()

    @classmethod
    def GetWallet1(cls, recreate=False):
        if cls._wallet1 is None or recreate:
            cls._wallet1 = UserWallet.Open(BoaTest.wallet_1_dest(), to_aes_key(BoaTest.wallet_1_pass()))
        return cls._wallet1


class BoaFixtureTest(WalletFixtureTestCase):

    dirname = None

    FIXTURE_REMOTE_LOC = 'https://s3.us-east-2.amazonaws.com/cityofzion/fixtures/fixtures_v7.tar.gz'
    FIXTURE_FILENAME = './fixtures/fixtures_v7.tar.gz'

    wallet_1_script_hash = UInt160(data=b'S\xefB\xc8\xdf!^\xbeZ|z\xe8\x01\xcb\xc3\xac/\xacI)')

    wallet_1_addr = 'APRgMZHZubii29UXF9uFa6sohrsYupNAvx'

    _wallet1 = None

    wallet_2_script_hash = UInt160(data=b'4\xd0=k\x80TF\x9e\xa8W\x83\xfa\x9eIv\x0b\x9bs\x9d\xb6')

    wallet_2_addr = 'ALb8FEhEmtSqv97fuNVuoLmcmrSKckffRf'

    _wallet2 = None

    wallet_3_script_hash = UInt160(data=b'\xa6\xc5\x9d\xeb\xf0\xd7(\xbd\x14\x89\xcd\xb9\xd9{\xd1\x90\xcb\x0b\xdch')

    wallet_3_addr = 'AWygZ1B5c3GDiLL6u5bHSVU45Ya1SVGX9P'

    _wallet3 = None

    @classmethod
    def setUpClass(cls):
        super(BoaFixtureTest, cls).setUpClass()
        cls.dirname = '/'.join(os.path.abspath(__file__).split('/')[:-3])

    @classmethod
    def GetWallet1(cls, recreate=False):
        if cls._wallet1 is None or recreate:
            cls._wallet1 = UserWallet.Open(BoaFixtureTest.wallet_1_dest(), to_aes_key(BoaFixtureTest.wallet_1_pass()))
        return cls._wallet1

    @classmethod
    def GetWallet2(cls, recreate=False):
        if cls._wallet2 is None or recreate:
            cls._wallet2 = UserWallet.Open(BoaFixtureTest.wallet_2_dest(), to_aes_key(BoaFixtureTest.wallet_2_pass()))
        return cls._wallet2

    @classmethod
    def GetWallet3(cls, recreate=False):
        if cls._wallet3 is None or recreate:
            cls._wallet3 = UserWallet.Open(BoaFixtureTest.wallet_3_dest(), to_aes_key(BoaFixtureTest.wallet_3_pass()))
        return cls._wallet3



from boa_test.tests.boa_test import BoaFixtureTest
from boa.compiler import Compiler
from neo.Core.TX.Transaction import Transaction
from neo.Prompt.Commands.BuildNRun import TestBuild

from neo.Prompt.Commands.BuildNRun import TestBuild
from neo.EventHub import events, SmartContractEvent


class TestContract(BoaFixtureTest):

    def test_Runtime(self):

        dispatched_events = []
        dispatched_logs = []

        def on_notif(evt):
            dispatched_events.append(evt)

        def on_log(evt):
            dispatched_logs.append(evt)
        events.on(SmartContractEvent.RUNTIME_NOTIFY, on_notif)
        events.on(SmartContractEvent.RUNTIME_LOG, on_log)

        output = Compiler.instance().load('%s/boa_test/example/blockchain/RuntimeTest.py' % TestContract.dirname).default
        out = output.write()

        tx, results, total_ops, engine = TestBuild(out, ['get_time', 1], self.GetWallet1(), '0202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b'\x01r\x04Z'))

        tx, results, total_ops, engine = TestBuild(out, ['check_witness', 1234], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBoolean(), False)

        tx, results, total_ops, engine = TestBuild(out, ['check_witness', bytearray(b'S\xefB\xc8\xdf!^\xbeZ|z\xe8\x01\xcb\xc3\xac/\xacI)')], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBoolean(), True)

        tx, results, total_ops, engine = TestBuild(out, ['log', 'hello'], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBoolean(), True)
        self.assertEqual(len(dispatched_logs), 1)
        self.assertEqual(dispatched_logs[0].event_payload[0], b'hello')

        tx, results, total_ops, engine = TestBuild(out, ['notify', 1234], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBoolean(), True)
        self.assertEqual(len(dispatched_events), 1)
        self.assertEqual(int.from_bytes(dispatched_events[0].event_payload[0], 'little'), 1234)

        tx, results, total_ops, engine = TestBuild(out, ['get_trigger', 1234], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 16)

        events.off(SmartContractEvent.RUNTIME_NOTIFY, on_notif)
        events.off(SmartContractEvent.RUNTIME_LOG, on_log)

    def test_Triggers(self):

        output = Compiler.instance().load('%s/boa_test/example/blockchain/TriggerTypeTest.py' % TestContract.dirname).default
        out = output.write()

        tx, results, total_ops, engine = TestBuild(out, [1], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b'\x10'))

        tx, results, total_ops, engine = TestBuild(out, [2], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b'\x00'))

        tx, results, total_ops, engine = TestBuild(out, [3], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b'\x20'))

        tx, results, total_ops, engine = TestBuild(out, [0], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), -1)



from boa_test.tests.boa_test import BoaTest
from boa.compiler import Compiler

from neo.Prompt.Commands.BuildNRun import TestBuild


class TestContract(BoaTest):

    def test_aWhile1(self):
        output = Compiler.instance().load('%s/boa_test/example/WhileTest1.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 6)

    def test_aWhile2(self):
        output = Compiler.instance().load('%s/boa_test/example/WhileTest2.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 6)

    def test_aWhile3(self):
        output = Compiler.instance().load('%s/boa_test/example/WhileTest.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 24)

    def test_Iter1(self):
        output = Compiler.instance().load('%s/boa_test/example/IterTest.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 18)

    def test_Iter2(self):
        output = Compiler.instance().load('%s/boa_test/example/IterTest2.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 8)

    def test_Iter3(self):
        output = Compiler.instance().load('%s/boa_test/example/IterTest3.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 7)

    def test_Iter4(self):
        output = Compiler.instance().load('%s/boa_test/example/IterTest4.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b'abcdabcdabcd\x0c'))

    def test_Iter5(self):
        output = Compiler.instance().load('%s/boa_test/example/IterTest5.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 51)

    def test_Range1(self):
        output = Compiler.instance().load('%s/boa_test/example/RangeTest.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        res = results[0].GetArray()
        self.assertEqual(len(res), 20)

    def test_Range2(self):
        output = Compiler.instance().load('%s/boa_test/example/IterTest6.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 10)

    def test_Range3(self):
        output = Compiler.instance().load('%s/boa_test/example/IterTest7.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 12)

    def test_Range4(self):
        output = Compiler.instance().load('%s/boa_test/example/IterTest8.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 6)





class list(list):

    """
    list() -> new empty list
    list(length=x) -> new list initialized with length of x
    """

#    def __init__(self, length=0):
#        pass
    pass


def concat(str1, str2):
    """
     range(str1, str2) -> str object

     Return a string that is the concatenation of the two arguments ( str1 + str2 )
     """
    pass


def keys(dictionary):
    """
     k = keys(mydict)

     pushes a list of a dictionary keys onto the stack
     """
    pass


def values(dictionary):
    """
     v = values(mydict)

     pushes a list of a dictionary values onto the stack
     """
    pass


def has_key(dictionary, key):
    """
     val = has_key(mydict, 'b')

     pushes a boolean of wether a dictionary has a key onto the stack
     """
    pass


# This is not necessary.  you can use mystring[start:end]
# Actually, it is currently necessary, as mystring[start:end] is not working as expected

def substr(source, start_index, count):
    """
    substr(source, start_index, count) -> list object

    Return a subset of a string `source`, starting at `start_index` and
    of length `count`
    """
    pass


def take(source, count):
    """
    take(source, count) -> list object

    Return a subset of a string or list `source`, starting
    at index 0 and of length `count`
    """
    pass


def range(start, stop):
    """
    range(start, stop) -> list object

    Return an list that is a a sequence of integers from start (inclusive)
    to stop (exclusive).  range(i, j) produces i, i+1, i+2, ..., j-1.
    """

    length = stop - start

    out = list(length=length)

    index = 0

    orig_start = start

    while start < stop:
        val = index + orig_start
        out[index] = val
        index = index + 1
        start = orig_start + index

    return out


def sha1(data):
    """

    :param data:
    """
    pass


def sha256(data):
    """

    :param data:
    """
    pass


def hash160(data):
    """

    :param data:
    """
    pass


def hash256(data):
    """

    :param data:
    """
    pass


def verify_signature(pubkey, signature, message):
    """
    :param pubkey:
    :param signature:
    :param message:
    """
    pass


def throw_if_null(item):
    pass


def breakpoint():
    """
    Adds a breakpoint to the debug map
    """
    pass



from boa.interop.System.ExecutionEngine import *


def Main(operation):

    if operation == 'executing_sh':

        return GetExecutingScriptHash()

    elif operation == 'script_container':
        return GetScriptContainer()

    elif operation == 'calling_sh':
        return GetCallingScriptHash()

    elif operation == 'entry_sh':
        return GetEntryScriptHash()

    return 'unknown'



from boa_test.tests.boa_test import BoaTest
from boa.compiler import Compiler
from neo.Prompt.Commands.BuildNRun import TestBuild


class TestContract(BoaTest):

    def test_list0(self):
        output = Compiler.instance().load('%s/boa_test/example/ArrayTest.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [0], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 1)

        tx, results, total_ops, engine = TestBuild(out, [1], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 6)

        tx, results, total_ops, engine = TestBuild(out, [2], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 3)

        tx, results, total_ops, engine = TestBuild(out, [4], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 8)

        tx, results, total_ops, engine = TestBuild(out, [8], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 9)

    def test_list1(self):
        output = Compiler.instance().load('%s/boa_test/example/ArrayTest1.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBoolean(), True)

    def test_list2(self):
        output = Compiler.instance().load('%s/boa_test/example/ArrayTest2.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b'\xa0'))

    def test_list3(self):
        output = Compiler.instance().load('%s/boa_test/example/ArrayTest3.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '02')
        self.assertEqual(len(results), 1)
        res = results[0].GetArray()
        self.assertEqual(len(res), 3)
        self.assertEqual(res[0].GetBigInteger(), 1)

    def test_list4(self):
        output = Compiler.instance().load('%s/boa_test/example/AppendTest.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '02')
        self.assertEqual(len(results), 1)
        res = results[0].GetArray()
        self.assertEqual(len(res), 2)
        self.assertEqual(res[0].GetBigInteger(), 6)

    def test_list5(self):
        output = Compiler.instance().load('%s/boa_test/example/ArrayRemoveTest.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '02')
        self.assertEqual(len(results), 1)
        res = results[0].GetArray()
        self.assertEqual(len(res), 3)
        self.assertEqual(res[0].GetBigInteger(), 16)
        self.assertEqual(res[1].GetBigInteger(), 3)

    def test_list6(self):
        output = Compiler.instance().load('%s/boa_test/example/ArrayReverseTest.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetString(), 'blah')

    def test_list7(self):
        output = Compiler.instance().load('%s/boa_test/example/ArrayTest4.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [], self.GetWallet1(), '', '02')
        self.assertEqual(len(results), 1)
        res = results[0].GetArray()
        self.assertEqual(len(res), 3)
        self.assertEqual(res[0].GetBigInteger(), 3)




class InvocationTransaction:

    @property
    def Script(self):
        """

        :return:
        """
        return GetScript(self)


def GetScript(invocation_transaction):
    """

    :param invocation_transaction:
    """
    pass





def GetHeight():
    """

    """
    pass


def GetHeader(height_or_hash):
    """

    :param height_or_hash:
    """
    pass


def GetBlock(height_or_hash):
    """

    :param height_or_hash:
    """
    pass


def GetTransaction(hash):
    """

    :param hash:
    """
    pass


def GetAccount(script_hash):
    """

    :param script_hash:
    """
    pass


def GetValidators():
    """

    """
    pass


def GetAsset(asset_id):
    """

    :param asset_id:
    """
    pass


def GetContract(script_hash):
    """

    :param script_hash:
    """
    pass




class Contract:

    @property
    def Script(self):
        """

        :return:
        """
        return GetScript(self)

    @property
    def StorageContext(self):
        """

        :return:
        """
        return GetStorageContext(self)


def GetScript(contract):
    """

    :param contract:
    """
    pass


def GetStorageContext(contract):
    """

    :param contract:
    """
    pass


def Create(script, parameter_list, return_type, properties, name, version, author, email, description):
    """

    :param script:
    :param parameter_list:
    :param return_type:
    :param properties:
    :param name
    :param version:
    :param author:
    :param email:
    :param description:
    """
    pass


def Migrate(script, parameter_list, return_type, properties, name, version, author, email, description):
    """

    :param script:
    :param parameter_list:
    :param return_type:
    :param need_storage:
    :param name
    :param version:
    :param author:
    :param email:
    :param description:
    """
    pass


def Destroy():
    """

    :param contract:
    """
    pass



from boa.interop.Neo.Blockchain import GetHeight
from boa.interop.Neo.Runtime import CheckWitness
from boa.interop.Neo.Action import RegisterAction
from boa.interop.Neo.Storage import Get, Put
from boa.builtins import concat
from boa_test.example.demo.nex.token import *
from boa_test.example.demo.nex.txio import get_asset_attachments

# OnInvalidKYCAddress = RegisterAction('invalid_registration', 'address')
OnKYCRegister = RegisterAction('kyc_registration', 'address')
OnTransfer = RegisterAction('transfer', 'addr_from', 'addr_to', 'amount')
OnRefund = RegisterAction('refund', 'addr_to', 'amount')


def kyc_register(ctx, args):
    """

    :param args:list a list of addresses to register
    :param token:Token A token object with your ICO settings
    :return:
        int: The number of addresses to register for KYC
    """
    ok_count = 0

    if CheckWitness(TOKEN_OWNER):

        for address in args:

            if len(address) == 20:

                kyc_storage_key = concat(KYC_KEY, address)
                Put(ctx, kyc_storage_key, True)

                OnKYCRegister(address)
                ok_count += 1

    return ok_count


def kyc_status(ctx, args):
    """
    Gets the KYC Status of an address

    :param args:list a list of arguments
    :return:
        bool: Returns the kyc status of an address
    """

    if len(args) > 0:
        addr = args[0]

        kyc_storage_key = concat(KYC_KEY, addr)

        return Get(ctx, kyc_storage_key)

    return False


def perform_exchange(ctx):
    """

     :param token:Token The token object with NEP5/sale settings
     :return:
         bool: Whether the exchange was successful
     """

    attachments = get_asset_attachments()  # [receiver, sender, neo, gas]

    # this looks up whether the exchange can proceed
    exchange_ok = can_exchange(ctx, attachments, False)

    if not exchange_ok:
        # This should only happen in the case that there are a lot of TX on the final
        # block before the total amount is reached.  An amount of TX will get through
        # the verification phase because the total amount cannot be updated during that phase
        # because of this, there should be a process in place to manually refund tokens
        if attachments[2] > 0:
            OnRefund(attachments[1], attachments[2])
        # if you want to exchange gas instead of neo, use this
        # if attachments.gas_attached > 0:
        #    OnRefund(attachments.sender_addr, attachments.gas_attached)
        return False

    # lookup the current balance of the address
    current_balance = Get(ctx, attachments[1])

    # calculate the amount of tokens the attached neo will earn
    exchanged_tokens = attachments[2] * TOKENS_PER_NEO / 100000000

    # if you want to exchange gas instead of neo, use this
    # exchanged_tokens += attachments[3] * TOKENS_PER_GAS / 100000000

    # add it to the the exchanged tokens and persist in storage
    new_total = exchanged_tokens + current_balance
    Put(ctx, attachments[1], new_total)

    # update the in circulation amount
    result = add_to_circulation(ctx, exchanged_tokens)

    # dispatch transfer event
    OnTransfer(attachments[0], attachments[1], exchanged_tokens)

    return True


def can_exchange(ctx, attachments, verify_only):
    """
    Determines if the contract invocation meets all requirements for the ICO exchange
    of neo or gas into NEP5 Tokens.
    Note: This method can be called via both the Verification portion of an SC or the Application portion

    When called in the Verification portion of an SC, it can be used to reject TX that do not qualify
    for exchange, thereby reducing the need for manual NEO or GAS refunds considerably

    :param attachments:Attachments An attachments object with information about attached NEO/Gas assets
    :return:
        bool: Whether an invocation meets requirements for exchange
    """

    # if you are accepting gas, use this
#        if attachments[3] == 0:
#            print("no gas attached")
#            return False

    # if youre accepting neo, use this

    if attachments[2] == 0:
        return False

    # the following looks up whether an address has been
    # registered with the contract for KYC regulations
    # this is not required for operation of the contract

#        status = get_kyc_status(attachments.sender_addr, storage)
    if not get_kyc_status(ctx, attachments[1]):
        return False

    # caluclate the amount requested
    amount_requested = attachments[2] * TOKENS_PER_NEO / 100000000

    # this would work for accepting gas
    # amount_requested = attachments.gas_attached * token.tokens_per_gas / 100000000

    exchange_ok = calculate_can_exchange(ctx, amount_requested, attachments[1], verify_only)

    return exchange_ok


def get_kyc_status(ctx, address):
    """
    Looks up the KYC status of an address

    :param address:bytearray The address to lookup
    :param storage:StorageAPI A StorageAPI object for storage interaction
    :return:
        bool: KYC Status of address
    """
    kyc_storage_key = concat(KYC_KEY, address)

    return Get(ctx, kyc_storage_key)


def calculate_can_exchange(ctx, amount, address, verify_only):
    """
    Perform custom token exchange calculations here.

    :param amount:int Number of tokens to convert from asset to tokens
    :param address:bytearray The address to mint the tokens to
    :return:
        bool: Whether or not an address can exchange a specified amount
    """
    height = GetHeight()

    current_in_circulation = Get(ctx, TOKEN_CIRC_KEY)

    new_amount = current_in_circulation + amount

    if new_amount > TOKEN_TOTAL_SUPPLY:
        return False

    if height < BLOCK_SALE_START:
        return False

    # if we are in free round, any amount
    if height > LIMITED_ROUND_END:
        return True

    # check amount in limited round
    if amount <= MAX_EXCHANGE_LIMITED_ROUND:

        # check if they have already exchanged in the limited round
        r1key = concat(address, LIMITED_ROUND_KEY)
        has_exchanged = Get(ctx, r1key)

        # if not, then save the exchange for limited round
        if not has_exchanged:
            # note that this method can be invoked during the Verification trigger, so we have the
            # verify_only param to avoid the Storage.Put during the read-only Verification trigger.
            # this works around a "method Neo.Storage.Put not found in ->" error in InteropService.py
            # since Verification is read-only and thus uses a StateReader, not a StateMachine
            if not verify_only:
                Put(ctx, r1key, True)
            return True

        return False

    return False



# tested

from boa.interop.Neo.Runtime import *


def Main(operation, arg):

    if operation == 'get_trigger':
        return GetTrigger()

    elif operation == 'check_witness':
        return CheckWitness(arg)

    elif operation == 'get_time':
        return GetTime()

    elif operation == 'log':
        Log(arg)
        return True

    elif operation == 'notify':
        Notify(arg)
        return True

    return 'unknown'



# tested

from boa.builtins import concat


def Main():

    items = [0, 1, 2]

    items2 = ['a', 'b', 'c', 'd']
    count = 0

    blah = b''

    for i in items:

        for j in items2:

            blah = concat(blah, j)

            count += 1

    blah = concat(blah, count)

    return blah



from boa_test.tests.boa_test import BoaTest
from boa.compiler import Compiler

from neo.Prompt.Commands.BuildNRun import TestBuild


class TestContract(BoaTest):

    def test_Throw1(self):
        output = Compiler.instance().load('%s/boa_test/example/ThrowTest.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [1], self.GetWallet1(), '02', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBoolean(), True)

        tx, results, total_ops, engine = TestBuild(out, [4], self.GetWallet1(), '02', '07')
        self.assertEqual(len(results), 0)

    def test_Throw2(self):
        output = Compiler.instance().load('%s/boa_test/example/ThrowIfNotTest.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [True], self.GetWallet1(), '01', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBoolean(), True)

        tx, results, total_ops, engine = TestBuild(out, [False], self.GetWallet1(), '01', '07')
        self.assertEqual(len(results), 0)



# tested

from boa.builtins import has_key


def Main():

    d = {
        'a': 1,
        'b': 4,
    }

    result = 0

    if has_key(d, 'a'):
        result += 2

    if not d.has_key('j'):
        result += 3

    if 'b' in d:

        print("b i si in d:")
        result += 17
    else:

        print("b not in d")

    return result  # should be 22



# tested

from boa.builtins import concat
from boa.interop.Neo.Runtime import Notify


def Main():

    items = [0, 1, 2]

    items2 = ['a', 'b', 'c', 'd']
    count = 0

    for i in items:  # 3

        count += 1

        for j in items2:  # 4

            count += 1

            for k in items:  # 3
                count += 1

    return count
